unit uSubstitution;
{$INCLUDE 'Plugin.inc'}

interface

uses
  Windows, SysUtils, Classes, Generics.Collections;

type
  TSubstitutionEngine = class;
  TSubstitutionEngineClass = class of TSubstitutionEngine;

  TSubstitutionEngines = class
  protected
    type TEngines = TObjectList<TSubstitutionEngine>;
  private
    fEngines: TEngines;
    function GetEngines: TEngines;
  protected
    property Engines: TEngines read GetEngines;
  public
    constructor Create;
    destructor Destroy; override;
    function AddEngine(Engine: TSubstitutionEngineClass): boolean; overload;
    function AddEngine(Engine: TSubstitutionEngine): boolean; overload;
    procedure SubstituteVariables(var Subject: string; var Position: integer); overload; virtual;
    function SubstituteVariables(const Subject: string): string; overload; virtual;
  end;

  TSubstitutionEngine = class
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function SubstituteVariable(var Subject: string; var Position: integer): boolean; virtual; abstract;
  end;

  TSimpleSubstitutionEngine = class(TSubstitutionEngine)
  protected
    function GetCaseSensitive: boolean; virtual;
    function GetPatternStart: string; virtual; abstract;
    function GetPatternEnd: string; virtual; abstract;
    function GetSubstitution(const Content: string; out NewContent: string): boolean; virtual; abstract;
    function PatternMatch(const Pattern, Subject: string; PatternLength, SubjectPosition, SubjectRemainingLength: integer; CaseSensitive: boolean): boolean;
  public
    function SubstituteVariable(var Subject: string; var Position: integer): boolean; override;
  end;

  TRegistrySubstitutionEngine = class(TSimpleSubstitutionEngine)
  // '$(reg:HKEY_LOCAL_MACHINE/Software/Microsoft/Windows/CurrentVersion/Run/SomeProgram)'
  protected
    function GetCaseSensitive: boolean; override;
    function GetPatternStart: string; override;
    function GetPatternEnd: string; override;
    function GetSubstitution(const Content: string; out NewContent: string): boolean; override;
  end;

  TEnvVarSubstitutionEngine = class(TSimpleSubstitutionEngine)
  // '$(env:USERPROFILE)'
  protected
    function GetCaseSensitive: boolean; override;
    function GetPatternStart: string; override;
    function GetPatternEnd: string; override;
    function GetSubstitution(const Content: string; out NewContent: string): boolean; override;
  end;

  TBatchEnvVarSubstitutionEngine = class(TSimpleSubstitutionEngine)
  // '%USERPROFILE%'
  protected
    function GetCaseSensitive: boolean; override;
    function GetPatternStart: string; override;
    function GetPatternEnd: string; override;
    function GetSubstitution(const Content: string; out NewContent: string): boolean; override;
  end;

  TKnownFolderSubstitutionEngine = class(TSimpleSubstitutionEngine)
  // '$(dir:STARTMENU)' -- comes from CSIDL_STARTMENU
  protected
    function GetCaseSensitive: boolean; override;
    function GetPatternStart: string; override;
    function GetPatternEnd: string; override;
    function GetSubstitution(const Content: string; out NewContent: string): boolean; override;
  end;

  TKnownFolderSubstitutionEngine2 = class(TKnownFolderSubstitutionEngine)
  // '$(CSIDL_STARTMENU)'
  protected
    function GetPatternStart: string; override;
  end;

implementation

uses
  Registry, ShlObj, uSystem;

{ TSubstitutionEngines }

function TSubstitutionEngines.AddEngine(Engine: TSubstitutionEngineClass): boolean;
var
  i: integer;
  Eng: TSubstitutionEngine;
begin
  for i := 0 to Pred(Engines.Count) do
  begin
    if Engines[i].ClassType = Engine then
    begin
      Result := False;
      Exit;
    end;
  end;
  Eng := Engine.Create;
  try
    Result := AddEngine(Eng);
    if Result then
      Eng := nil;
  finally
    FreeAndNil(Eng);
  end;
end;

function TSubstitutionEngines.AddEngine(Engine: TSubstitutionEngine): boolean;
begin
  Result := False;
  if Engines.IndexOf(Engine) < 0 then
  begin
    Engines.Add(Engine);
    Result := True;
  end;
end;

constructor TSubstitutionEngines.Create;
begin
  inherited Create;
  fEngines := nil;
end;

destructor TSubstitutionEngines.Destroy;
begin
  FreeAndNil(fEngines);
  inherited;
end;

function TSubstitutionEngines.GetEngines: TEngines;
begin
  if fEngines = nil then
  begin
    fEngines := TEngines.Create;
  end;
  Result := fEngines;
end;

function TSubstitutionEngines.SubstituteVariables(const Subject: string): string;
var
  Index, LastIndex: integer;
begin
  Result := Subject;
  if fEngines <> nil then
  begin
    Index := 1;
    while Index <= Length(Result) do
    begin
      LastIndex := Index;
      SubstituteVariables(Result, Index);
      if Index = LastIndex then
        Inc(Index);
    end;
  end;
end;

procedure TSubstitutionEngines.SubstituteVariables(var Subject: string; var Position: integer);
var
  i: integer;
begin
  if fEngines <> nil then
  begin
    for i := 0 to Pred(Engines.Count) do
    begin
      if Engines[i].SubstituteVariable(Subject, Position) then
        Break;
    end;
  end;
end;

{ TSubstitutionEngine }

constructor TSubstitutionEngine.Create;
begin
  inherited Create;
end;

destructor TSubstitutionEngine.Destroy;
begin
  inherited;
end;

{ TSimpleSubstitutionEngine }

function TSimpleSubstitutionEngine.GetCaseSensitive: boolean;
begin
  Result := True;
end;

function TSimpleSubstitutionEngine.PatternMatch(const Pattern, Subject: string; PatternLength, SubjectPosition, SubjectRemainingLength: integer; CaseSensitive: boolean): boolean;
var
  i: integer;
  Match: boolean;
begin
  Result := False;
  if (PatternLength > 0) and (SubjectRemainingLength >= PatternLength) then
  begin
    for i := 1 to PatternLength do
    begin
      if CaseSensitive then
        Match := Pattern[i] = Subject[i + SubjectPosition - 1]
      else
        Match := CharUpper(Pointer(Pattern[i])) = CharUpper(Pointer(Subject[i + SubjectPosition - 1]));
      if not Match then
        Exit;
    end;
    Result := True;
  end;
end;

function TSimpleSubstitutionEngine.SubstituteVariable(var Subject: string; var Position: integer): boolean;
var
  CaseSensitive: boolean;
  PatternStart, PatternEnd, Content, NewContent: string;
  SubjectRemainingLength, PatternStartLength, PatternEndLength: integer;
  ContentStart, ContentEnd, ContentLength: integer;
begin
  Result := False;
  SubjectRemainingLength := Length(Subject) - Position + 1;
  if SubjectRemainingLength <= 0 then
    Exit;
  PatternStart := Self.GetPatternStart;
  PatternStartLength := Length(PatternStart);
  if PatternStartLength <= 0 then
    Exit;
  PatternEnd := Self.GetPatternEnd;
  PatternEndLength := Length(PatternEnd);
  if PatternEndLength <= 0 then
    Exit;
  CaseSensitive := Self.GetCaseSensitive;
  if PatternMatch(PatternStart, Subject, PatternStartLength, Position, SubjectRemainingLength, CaseSensitive) then
  begin
    ContentStart := Position + PatternStartLength;
    ContentEnd := ContentStart;
    SubjectRemainingLength := SubjectRemainingLength - PatternStartLength;
    while (SubjectRemainingLength >= PatternEndLength) do
    begin
      if PatternMatch(PatternEnd, Subject, PatternEndLength, ContentEnd, SubjectRemainingLength, CaseSensitive) then
      begin
        ContentLength := ContentEnd - ContentStart;
        Content := Copy(Subject, ContentStart, ContentLength);
        if GetSubstitution(Content, NewContent) then
        begin
          System.Delete(Subject, Position, PatternStartLength + ContentLength + PatternEndLength);
          System.Insert(NewContent, Subject, Position);
          Position := Position + Length(NewContent);
          Result := True;
          Break;
        end;
      end;
      Inc(ContentEnd);
      Dec(SubjectRemainingLength);
    end;
  end;
end;

{ TRegistrySubstitutionEngine }

function TRegistrySubstitutionEngine.GetCaseSensitive: boolean;
begin
  Result := False;
end;

function TRegistrySubstitutionEngine.GetPatternEnd: string;
begin
  Result := ')';
end;

function TRegistrySubstitutionEngine.GetPatternStart: string;
begin
  Result := '$(reg:';
end;

function TRegistrySubstitutionEngine.GetSubstitution(const Content: string; out NewContent: string): boolean;
var
  i, Index: integer;
  sContent, sRoot, sKey, sValue: string;
  Reg: TRegistry;
  RootKey: HKEY;
begin
  Result := False;
  if Content <> '' then
  begin
    sContent := Content;
    for i := 1 to Length(Content) do
    begin
      if (sContent[i] = '\') or (sContent[i] = '/') then
        sContent[i] := PathDelim;
    end;
    RootKey := HKEY_CURRENT_USER;
    Index := Pos(PathDelim, sContent);
    if Index > 0 then
    begin
      sRoot := AnsiUpperCase(Copy(sContent, 1, Pred(Index)));
      sKey := Copy(sContent, Succ(Index), MaxInt);
      if sRoot = 'HKEY_CURRENT_USER' then RootKey := HKEY_CURRENT_USER
      else if sRoot = 'HKCU' then RootKey := HKEY_CURRENT_USER
      else if sRoot = 'HKEY_LOCAL_MACHINE' then RootKey := HKEY_LOCAL_MACHINE
      else if sRoot = 'HKLM' then RootKey := HKEY_LOCAL_MACHINE
      else if sRoot = 'HKEY_CLASSES_ROOT' then RootKey := HKEY_CLASSES_ROOT
      else if sRoot = 'HKCR' then RootKey := HKEY_CLASSES_ROOT
      else if sRoot = 'HKEY_USERS' then RootKey := HKEY_USERS
      else if sRoot = 'HKU' then RootKey := HKEY_USERS
      else if sRoot = 'HKEY_PERFORMANCE_DATA' then RootKey := HKEY_PERFORMANCE_DATA
      else if sRoot = 'HKPD' then RootKey := HKEY_PERFORMANCE_DATA
      else if sRoot = 'HKEY_CURRENT_CONFIG' then RootKey := HKEY_CURRENT_CONFIG
      else if sRoot = 'HKCC' then RootKey := HKEY_CURRENT_CONFIG
      else if sRoot = 'HKEY_DYN_DATA' then RootKey := HKEY_DYN_DATA
      else if sRoot = 'HKDD' then RootKey := HKEY_DYN_DATA
      else ;
    end
    else
    begin
      sKey := sContent;
    end;
    sValue := ExtractFileName(sKey);
    sKey := ExcludeTrailingPathDelimiter(ExtractFilePath(sKey));
    Reg := TRegistry.Create(KEY_READ);
    try
      Reg.RootKey := RootKey;
      if Reg.OpenKeyReadOnly(sKey) then
      begin
        if Reg.ValueExists(sValue) then
        begin
          NewContent := Reg.ReadString(sValue);
          Result := True;
        end;
      end;
    finally
      FreeAndNil(Reg);
    end;
  end;
end;

{ TEnvVarSubstitutionEngine }

function TEnvVarSubstitutionEngine.GetCaseSensitive: boolean;
begin
  Result := False;
end;

function TEnvVarSubstitutionEngine.GetPatternEnd: string;
begin
  Result := ')';
end;

function TEnvVarSubstitutionEngine.GetPatternStart: string;
begin
  Result := '$(env:';
end;

function TEnvVarSubstitutionEngine.GetSubstitution(const Content: string; out NewContent: string): boolean;
begin
  if Content <> '' then
  begin
    NewContent := GetEnvironmentVariable(Content);
    Result := True;
  end
  else
  begin
    Result := False;
  end;
end;

{ TBatchEnvVarSubstitutionEngine }

function TBatchEnvVarSubstitutionEngine.GetCaseSensitive: boolean;
begin
  Result := False;
end;

function TBatchEnvVarSubstitutionEngine.GetPatternEnd: string;
begin
  Result := '%';
end;

function TBatchEnvVarSubstitutionEngine.GetPatternStart: string;
begin
  Result := '%';
end;

function TBatchEnvVarSubstitutionEngine.GetSubstitution(const Content: string; out NewContent: string): boolean;
begin
  if Content = '' then
    NewContent := '%'
  else
    NewContent := GetEnvironmentVariable(Content);
  Result := True;
end;

{ TKnownFolderSubstitutionEngine }

function TKnownFolderSubstitutionEngine.GetCaseSensitive: boolean;
begin
  Result := False;
end;

function TKnownFolderSubstitutionEngine.GetPatternEnd: string;
begin
  Result := ')';
end;

function TKnownFolderSubstitutionEngine.GetPatternStart: string;
begin
  Result := '$(dir:';
end;

function TKnownFolderSubstitutionEngine.GetSubstitution(const Content: string; out NewContent: string): boolean;
var
  sContent: string;
  FolderID: integer;
begin
  Result := False;
  if Content <> '' then
  begin
    sContent := AnsiUpperCase(Content);
    if      sContent = 'DESKTOP' then FolderID := CSIDL_DESKTOP
    else if sContent = 'INTERNET' then FolderID := CSIDL_INTERNET
    else if sContent = 'PROGRAMS' then FolderID := CSIDL_PROGRAMS
    else if sContent = 'CONTROLS' then FolderID := CSIDL_CONTROLS
    else if sContent = 'PRINTERS' then FolderID := CSIDL_PRINTERS
    else if sContent = 'PERSONAL' then FolderID := CSIDL_PERSONAL
    else if sContent = 'FAVORITES' then FolderID := CSIDL_FAVORITES
    else if sContent = 'STARTUP' then FolderID := CSIDL_STARTUP
    else if sContent = 'RECENT' then FolderID := CSIDL_RECENT
    else if sContent = 'SENDTO' then FolderID := CSIDL_SENDTO
    else if sContent = 'BITBUCKET' then FolderID := CSIDL_BITBUCKET
    else if sContent = 'STARTMENU' then FolderID := CSIDL_STARTMENU
    else if sContent = 'MYMUSIC' then FolderID := CSIDL_MYMUSIC
    else if sContent = 'MYVIDEO' then FolderID := CSIDL_MYVIDEO
    else if sContent = 'DESKTOPDIRECTORY' then FolderID := CSIDL_DESKTOPDIRECTORY
    else if sContent = 'DRIVES' then FolderID := CSIDL_DRIVES
    else if sContent = 'NETWORK' then FolderID := CSIDL_NETWORK
    else if sContent = 'NETHOOD' then FolderID := CSIDL_NETHOOD
    else if sContent = 'FONTS' then FolderID := CSIDL_FONTS
    else if sContent = 'TEMPLATES' then FolderID := CSIDL_TEMPLATES
    else if sContent = 'COMMON_STARTMENU' then FolderID := CSIDL_COMMON_STARTMENU
    else if sContent = 'COMMON_PROGRAMS' then FolderID := CSIDL_COMMON_PROGRAMS
    else if sContent = 'COMMON_STARTUP' then FolderID := CSIDL_COMMON_STARTUP
    else if sContent = 'COMMON_DESKTOPDIRECTORY' then FolderID := CSIDL_COMMON_DESKTOPDIRECTORY
    else if sContent = 'APPDATA' then FolderID := CSIDL_APPDATA
    else if sContent = 'PRINTHOOD' then FolderID := CSIDL_PRINTHOOD
    else if sContent = 'LOCAL_APPDATA' then FolderID := CSIDL_LOCAL_APPDATA
    else if sContent = 'ALTSTARTUP' then FolderID := CSIDL_ALTSTARTUP
    else if sContent = 'COMMON_ALTSTARTUP' then FolderID := CSIDL_COMMON_ALTSTARTUP
    else if sContent = 'COMMON_FAVORITES' then FolderID := CSIDL_COMMON_FAVORITES
    else if sContent = 'INTERNET_CACHE' then FolderID := CSIDL_INTERNET_CACHE
    else if sContent = 'COOKIES' then FolderID := CSIDL_COOKIES
    else if sContent = 'HISTORY' then FolderID := CSIDL_HISTORY
    else if sContent = 'COMMON_APPDATA' then FolderID := CSIDL_COMMON_APPDATA
    else if sContent = 'WINDOWS' then FolderID := CSIDL_WINDOWS
    else if sContent = 'SYSTEM' then FolderID := CSIDL_SYSTEM
    else if sContent = 'PROGRAM_FILES' then FolderID := CSIDL_PROGRAM_FILES
    else if sContent = 'MYPICTURES' then FolderID := CSIDL_MYPICTURES
    else if sContent = 'PROFILE' then FolderID := CSIDL_PROFILE
    else if sContent = 'SYSTEMX86' then FolderID := CSIDL_SYSTEMX86
    else if sContent = 'PROGRAM_FILESX86' then FolderID := CSIDL_PROGRAM_FILESX86
    else if sContent = 'PROGRAM_FILES_COMMON' then FolderID := CSIDL_PROGRAM_FILES_COMMON
    else if sContent = 'PROGRAM_FILES_COMMONX86' then FolderID := CSIDL_PROGRAM_FILES_COMMONX86
    else if sContent = 'COMMON_TEMPLATES' then FolderID := CSIDL_COMMON_TEMPLATES
    else if sContent = 'COMMON_DOCUMENTS' then FolderID := CSIDL_COMMON_DOCUMENTS
    else if sContent = 'COMMON_ADMINTOOLS' then FolderID := CSIDL_COMMON_ADMINTOOLS
    else if sContent = 'ADMINTOOLS' then FolderID := CSIDL_ADMINTOOLS
    else if sContent = 'CONNECTIONS' then FolderID := CSIDL_CONNECTIONS
    else if sContent = 'COMMON_MUSIC' then FolderID := CSIDL_COMMON_MUSIC
    else if sContent = 'COMMON_PICTURES' then FolderID := CSIDL_COMMON_PICTURES
    else if sContent = 'COMMON_VIDEO' then FolderID := CSIDL_COMMON_VIDEO
    else if sContent = 'RESOURCES' then FolderID := CSIDL_RESOURCES
    else if sContent = 'RESOURCES_LOCALIZED' then FolderID := CSIDL_RESOURCES_LOCALIZED
    else if sContent = 'COMMON_OEM_LINKS' then FolderID := CSIDL_COMMON_OEM_LINKS
    else if sContent = 'CDBURN_AREA' then FolderID := CSIDL_CDBURN_AREA
    else if sContent = 'COMPUTERSNEARME' then FolderID := CSIDL_COMPUTERSNEARME
    else FolderID := -1;
    if FolderID >= 0 then
    begin
      NewContent := GetSpecialFolder(FolderID);
      Result := NewContent <> '';
    end;
  end;
end;

{ TKnownFolderSubstitutionEngine2 }

function TKnownFolderSubstitutionEngine2.GetPatternStart: string;
begin
  Result := '$(CSIDL_';
end;

end.
